# -*- coding: utf-8 -*-
#  ECS ComposeX <https://github.com/lambda-my-aws/ecs_composex>
#  Copyright (C) 2020  John Mille <john@lambda-my-aws.io>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Functions that will add networking elements for the ECS Service as well as generating
the network configuration for the ServiceDefinition
"""

from troposphere import Ref, Sub, GetAtt, Tags, If
from troposphere.ec2 import SecurityGroup
from troposphere.ecs import ServiceRegistry
from troposphere.servicediscovery import (
    DnsConfig as SdDnsConfig,
    Service as SdService,
    DnsRecord as SdDnsRecord,
    HealthCheckCustomConfig as SdHealthCheckCustomConfig,
)

from ecs_composex.common import KEYISSET, LOG
from ecs_composex.common.cfn_params import ROOT_STACK_NAME_T
from ecs_composex.ecs.ecs_loadbalancing import define_service_load_balancing
from ecs_composex.ecs.ecs_params import (
    SERVICE_NAME_T,
    SERVICE_NAME,
    SG_T,
    SERVICE_HOSTNAME,
)
from ecs_composex.ecs.ecs_conditions import USE_HOSTNAME_CON_T
from ecs_composex.vpc.vpc_conditions import USE_VPC_MAP_ID_CON_T
from ecs_composex.vpc.vpc_params import VPC_ID, VPC_MAP_ID


def add_service_default_sg(template):
    """
    Adds a default security group for the microservice.

    :param template: the template to add the SG to.
    :type template: troposphere.Template

    :return: Security group
    :type: troposphere.ec2.SecurityGroup
    """
    sg = SecurityGroup(
        SG_T,
        template=template,
        GroupDescription=Sub(
            f"SG for ${{{SERVICE_NAME_T}}} - ${{{ROOT_STACK_NAME_T}}}"
        ),
        Tags=Tags(
            {
                "Name": Sub(f"${{{SERVICE_NAME_T}}}-${{{ROOT_STACK_NAME_T}}}"),
                "StackName": Ref("AWS::StackName"),
                "MicroserviceName": Ref(SERVICE_NAME),
            }
        ),
        VpcId=Ref(VPC_ID),
    )
    return sg


def add_service_to_map(template, service_name, network_settings):
    """
    Function to add the service into a cloudmap
    """
    sd_service = SdService(
        "EcsDiscoveryService",
        template=template,
        Condition=USE_VPC_MAP_ID_CON_T,
        Description=f"{service_name}",
        NamespaceId=Ref(VPC_MAP_ID),
        HealthCheckCustomConfig=SdHealthCheckCustomConfig(FailureThreshold=1.0),
        DnsConfig=SdDnsConfig(
            RoutingPolicy="MULTIVALUE",
            NamespaceId=Ref("AWS::NoValue"),
            DnsRecords=[
                SdDnsRecord(TTL="30", Type="A"),
                SdDnsRecord(TTL="30", Type="SRV"),
            ],
        ),
        Name=If(USE_HOSTNAME_CON_T, Ref(SERVICE_HOSTNAME), Ref(SERVICE_NAME)),
    )
    registries = []
    for port in network_settings["ports"]:
        registry = ServiceRegistry(
            f"ServiceRegistry{port['published']}",
            RegistryArn=GetAtt(sd_service, "Arn"),
            Port=port["published"],
        )
        registries.append(registry)
    return registries


def define_protocol(port_string):
    """Function to define the port protocol. Default to TCP if not specified otherwise

    :param port_string: the port string to parse from the ports list in the compose file
    :type port_string: str

    :return: protocol, ie. udp or tcp
    :rtype: str
    """
    protocols = ["tcp", "udp"]
    protocol = "tcp"
    if port_string.find("/"):
        protocol_found = port_string.split("/")[-1].strip()
        if protocol_found in protocols:
            return protocol_found
    return protocol


def define_service_ports(service):
    """Function to define common structure to ports

    :param service: the service as defined in compose file
    :type service: dict

    :return: list of ports the service uses formatted according to dict
    :rtype: list
    """
    ports = []
    valid_keys = ["published", "target", "protocol", "mode"]
    service_ports = []
    if KEYISSET("ports", service):
        ports = service["ports"]
    for port in ports:
        if not isinstance(port, (str, dict, int)):
            raise TypeError(
                "ports must be of types", dict, "or", list, "got", type(port)
            )
        if isinstance(port, str):
            service_ports.append(
                {
                    "protocol": define_protocol(port),
                    "published": port.split(":")[0],
                    "target": port.split(":")[-1].split("/")[0].strip(),
                    "mode": "awsvpc",
                }
            )
        elif isinstance(port, dict):
            if not set(port).issubset(valid_keys):
                raise KeyError(f"Valid keys are", valid_keys, "got", port.keys())
            port["mode"] = "awsvpc"
            service_ports.append(port)
        elif isinstance(port, int):
            service_ports.append(
                {"protocol": "tcp", "published": port, "target": port, "mode": "awsvpc"}
            )
    LOG.debug(service_ports)
    return service_ports


def compile_network_settings(compose_content, service, service_name):
    """Function to generate a set of values used for all network related stuff

    :param compose_content: Docker/ComposeX File content
    :type compose_content: dict
    :param service: service definition as imported from Docker/ComposeX File
    :type service_name: dict
    :param service_name: name of the service as defined in Docker/ComposeX File
    :type service_name: str

    :return: settings
    :rtype: dict
    """
    valid_keys = [
        "use_alb",
        "use_nlb",
        "use_cloudmap",
        "is_public",
        "healthcheck",
        "ext_sources",
        "ports",
    ]
    settings = {}
    defaults = {
        "use_cloudmap": True,
        "use_nlb": False,
        "use_alb": False,
        "is_public": False,
    }
    if KEYISSET("labels", service):
        settings = service["labels"]
    if KEYISSET("configs", compose_content):
        if KEYISSET(service_name, compose_content["configs"]):
            LOG.warning(
                f"Service {service_name} has defined configs in the configs section."
                "Overriding with values"
            )
            settings = compose_content["configs"][service_name]
        if KEYISSET("globals", compose_content["configs"]) and KEYISSET(
            "network", compose_content["configs"]["globals"]["network"]
        ):
            defaults.update(compose_content["configs"]["globals"]["network"])

    if not set(settings).issubset(valid_keys):
        LOG.error(valid_keys)
        LOG.error(settings)
        raise KeyError(
            "Invalid keys defined in the network section of the service config in configs"
        )
    for key in valid_keys:
        if key not in settings and key != "ports":
            if key in defaults.keys():
                if key == "use_cloudmap" and KEYISSET("is_public", settings):
                    LOG.debug(f"Missing use_cloudmap but {service_name} is public.")
                    settings[key] = False
                else:
                    settings[key] = defaults[key]
    settings["ports"] = define_service_ports(service)
    LOG.debug(service_name)
    LOG.debug(settings)
    return settings


def define_service_network_config(template, service_name, network_settings, **kwargs):
    """Function to define microservice ingress.

    :param template: service template to add resources to
    :type template: troposphere.Template
    :param service_name: name of the service as defined in the compose file
    :type service_name: str
    :param network_settings: see :func:`ecs_composex.ecs.ecs_networking.compile_network_settings`
    :type network_settings: dict
    :param kwargs: Optional additional parameters

    :return: tuple of the args for the Service() and the dependencies
    :rtype: tuple
    """
    add_service_default_sg(template)
    service_lbs = Ref("AWS::NoValue")
    registries = add_service_to_map(template, service_name, network_settings)
    if not registries:
        registries = Ref("AWS::NoValue")
    service_attrs = {"LoadBalancers": service_lbs, "ServiceRegistries": registries}
    external_dependencies = []
    if not KEYISSET("AwsAzs", kwargs):
        raise KeyError(
            "Missing AwsAzs from options."
            "AZs are required to configure services networking"
        )
    if not KEYISSET("ports", network_settings):
        LOG.debug(f"{service_name} does not have any ports. No ingress necessary")
        return service_attrs, external_dependencies
    LOG.debug(network_settings)
    if KEYISSET("use_nlb", network_settings) or KEYISSET("use_alb", network_settings):
        service_lb = define_service_load_balancing(
            template, service_name, network_settings, **kwargs
        )
        service_attrs["LoadBalancers"] = service_lb[0]
        service_attrs["DependsOn"] = (
            service_lb[-1] if isinstance(service_lb[-1], list) else []
        )
    return service_attrs, external_dependencies
